(** Implémentation des flots. *)

type 'a t =
  'a head Lazy.t
and
  'a head =
    | Nil
    | Cons of 'a * 'a t

(** Implémentation  (polymorphe) du type des flots. 
  Une liste est entièrement stockée en mémoire dès sa création et donc avant 
  même que l'on y accède. À l'inverse, un flot peut être vu comme  
  une liste « potentielle », c'est-à-dire que ses éléments sont générés 
  « à la demande » sans être stockés en mémoire « à l'avance ». 
  Le type ['a t] est défini comme dans le manuel de cours.
*)

(** {2 Égalité} *)

(** [equal eq xs ys] détermine si les flots [xs] et [ys] contiennent les 
mêmes éléments au regard de la fonction [eq] qui teste l'égalité entre éléments. 
Les flots sont donc forcés jusqu'à la première différence. Peut ne pas 
terminer si les deux arguments sont infinis.*)
val equal: ('a -> 'b -> bool) -> 'a t -> 'b t -> bool

(** {2 Constructeurs} *)

(** [from_list l] renvoie le flot dont les éléments sont ceux de [l] *)
val from_list: 'a list -> 'a t

(** [unfold f seed] produit un flot (potentiellement infini) obtenu en itérant la 
fonction [f] à partir de la valeur initiale [seed]. À chaque requête, [f] peut 
ne rien renvoyer (valeur [None]) auquel cas le flot est épuisé, ou bien 
une paire de valeurs constituée du premier élément du flot et d'une valeur à partir de laquelle 
[f] pourra déterminer l'élément suivant du flot. 
Cette fonction est typiquement utilisée pour définir le flot associé à un 
fichier. Ce flot renvoie les octets contenus dans le fichier au fur et à mesure qu'il est sollicité.
*)
val unfold: ('b -> ('a * 'b) option) -> 'b -> 'a t


(** {2 Concaténation} *)

(** [concat xs ys] renvoie la concaténation des flots [xs] et [ys], donc un flot 
qui, lorsqu'il est sollicité, renvoie les éléments de [xs] puis, lorsque ce dernier est épuisé, 
ceux de [ys].
 
Si [xs] est infini, les flots [xs] et [concat xs ys] sont « égaux » au 
sens où les calculs effectués lorsque l'on force [concat xs ys] sont les mêmes que 
ceux qui seraient effectués si l'on forçait [xs].*)
val concat: 'a t -> 'a t -> 'a t


(* [flatten xss] renvoie la concaténation de tous les éléments du flot [xss] (puisque ce sont eux-même des flots).
 
Si l'un de éléments de [xss] est un flot infini, alors [flatten xss] est un flot infini.
Si [xss] est le flot infini des flots vide, alors [Lazy.force (flatten xss)] ne termine pas.*)
(*
val flatten: 'a t t -> 'a t
*)



(** {2 Préfixe fini} *)

(** [take n xs] renvoie le flot des [n] premiers éléments of [xs]. Si [xs] a 
moins de [n] éléments, alors les calculs effectués lorsque l'on force [take n 
xs] sont les mêmes que ceux qui seraient effectués si l'on forçait [xs]. *)
val take: int -> 'a t -> 'a t

(** [last xs] retourne [Some x] où [x] est le dernier élément d'un flot fini 
qui n'est pas vide, [None] si le flot est vide. Le calcul « épuise » son argument 
et donc ne termine pas si ce dernier est infini.*)
val last: 'a t -> 'a option

(** {2 Réducteurs} *)

(** [iter f xs] force [xs] pour obtenir son premier élément [x] et le reste 
[xs'] du flot, puis effectue [f x] et réitère en appelant [iter f xs']. 
En particulier [iter f xs] ne s'arrête que lorsque le flot [xs] est « épuisé » 
et peut donc ne pas terminer. *)
val iter: ('a -> unit) -> 'a t -> unit

(** [map f xs] applique la fonction [f] à chacun des éléments du flot [xs] pour 
créer un nouveau flot.*)
val map: ('a -> 'b) -> ('a t) -> 'b t

(* BEGIN SOLUTION *)

(* Suppose [f] is a function that consumes one element [x] of type ['a] and
   produces an arbitrary number of elements of type ['b], by concatenating
   them in front of its second argument. Then, [map_one_many f xs accu]
   consumes an ['a stream] to produce a ['b stream], by iterating the action
   of [f]. *)

(* Supposons que la fonction [f : 'a -> 'b t -> 'b t] ait pour effet 
d'ajouter, selon son premier argument, des éléments « à gauche » de son second argument. 

La fonction [map_one_many] vérifie la propriété suivante : si 
la fonction [f] se comporte comme la fonction [fun x accu -> concat (g x) accu] 
pour une certaine fonction [g : 'a -> 'b t] et que le flot [xs] est 
(x{_0}x{_1}x{_2}...)
alors 
[map_one_many f seed] 
doit renvoyer la concaténation des flots (g (x{_0})), (g (x{_1})), (g (x{_2})), ... et seed. 
Noter que si le flot [xs] est infini et que pour chacun de ses éléments [x], 
le flot [g x] est vide, alors l'expression [force (map_one_many f xs seed)] 
déclenche un calcul qui ne termine pas (alors que l'on pourrait penser que 
le flot obtenu est « équivalent » à [seed]). Un exemple d'utilisation 
typique : [xs] est un flot d'entier et [f] est une fonction qui ajoute le flot 
des bits d'un entier (son premier argument) 
« devant » son second argument (qui est donc un flot de bits).
*)
(** Voir la page du {{:../../INF441-TD8-1.html}TD8}*)
val map_one_many: ('a -> 'b t -> 'b t) -> 'a t -> (*'b t ->*) 'b t

(* La fonction [map_many_one] vérifie la propriété suivante : si la fonction 
[f] consomme un « segment initial » du flot [xs : 'a t] pour produire une 
valeur de type ['b] et le flot restant (i.e. le « segment terminal » des valeurs qui n'ont pas 
été consommées), alors la fonction [map_many_to_one f xs] doit renvoyer le flot 
des valeurs produites par les appels successifs de la fonction [f]. Un exemple 
d'utilisation typique : [xs] est un flot de bits et la fonction [f] consomme 
les {i k} premiers bits du flot pour produire un entier représentable sur {i k} bits. 
Plus généralement, c'est le principe de fonctionnement d'un {i lexer}. 
*)
(** Voir la page du {{:../../INF441-TD8-1.html}TD8}*)
val map_many_one: ('a t -> 'b * 'a t) -> 'a t -> 'b t

(* END SOLUTION *)

(** Retour à la page du {{:../INF441-TD8-1.html}TD8}*)
